REM
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM    *  BasicNews, a very simple news server for Linux.            *
REM    *  Copyright (C) Mateusz Viste 2008, 2009                     *
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM
REM  Written in FreeBASIC v0.20.0 by Mateusz Viste <mateusz@viste-family.net>
REM
REM   ----------------------------------------------------------------------
REM    This program is free software: you can redistribute it and/or modify
REM    it under the terms of the GNU General Public License as published by
REM    the Free Software Foundation, either version 3 of the License, or
REM    (at your option) any later version.
REM
REM    This program is distributed in the hope that it will be useful,
REM    but WITHOUT ANY WARRANTY; without even the implied warranty of
REM    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
REM    GNU General Public License for more details.
REM
REM    You should have received a copy of the GNU General Public License
REM    along with this program.  If not, see <http://www.gnu.org/licenses/>.
REM   ----------------------------------------------------------------------
REM

CONST pVer AS STRING = "0.13"
CONST pDate AS STRING = "2009"
CONST homepage AS STRING = "http://www.viste-family.net/mateusz/software/basicnews/"

CONST CRLF AS STRING = CHR(13) + CHR(10)
TYPE NewsGroup
  Name AS STRING
  Description AS STRING
  FirstArticle AS INTEGER
  LastArticle AS INTEGER
  ArticlesCount AS INTEGER
  WriteAccess AS BYTE
  Auth AS BYTE
  CurArticle AS INTEGER
  MaxArticleSize AS INTEGER
  CreationTime AS STRING
END TYPE
DIM SHARED AS STRING ConfigFile, LogFile, SpoolPath, LineBuff, CurUser, CurHost, AuthUser, Banner, ServerFQDN
DIM SHARED LastCommand(1 TO 10) AS STRING
DIM SHARED AS INTEGER VerboseMode, UserBaseSize, GroupCount
DIM SHARED AS BYTE WriteAccess, GlobalAuth
DIM SHARED CurGroup AS NewsGroup
DIM SHARED AS DOUBLE StartTime
REDIM SHARED UserBase(1 TO 4) AS STRING
REDIM SHARED GroupList(1 TO 2) AS STRING
StartTime = TIMER

#INCLUDE ONCE "functions.bi"     ' Include all needed subs & functions

IF LCASE(COMMAND(1)) = "-h" OR LCASE(COMMAND(1)) = "--help" THEN About()

ConfigFile = "/etc/basicnews.cfg"
LogFile = "/var/log/basicnews.log"
SpoolPath = ReadCFG(ConfigFile, "SpoolPath")
Banner = ReadCFG(ConfigFile, "Banner")
ServerFQDN = ReadCFG(ConfigFile, "ServerFQDN")
IF LEN(ServerFQDN) = 0 THEN
  ServerFQDN = "BasicNews.localhost"
  LogLine("No FQDN found! Please configure your FQDN in the BasicNews configuration file.")
END IF
WriteAccess = VAL(ReadCFG(ConfigFile, "WriteAccess"))
GlobalAuth = VAL(ReadCFG(ConfigFile, "GlobalAuth"))
VerboseMode = VAL(ReadCFG(ConfigFile, "VerboseMode"))
IF LEN(Banner) = 0 THEN Banner = "Welcome to this BasicNews server!"
IF LEN(SpoolPath) = 0 THEN SpoolPath = "/var/basicnews/"
IF RIGHT(SpoolPath, 1) <> "/" THEN SpoolPath += "/"

IF LEN(ENVIRON("REMOTE_ADDR")) > 0 THEN CurHost = ENVIRON("REMOTE_ADDR")    '  Get remote
IF LEN(ENVIRON("REMOTE_HOST")) > 0 THEN CurHost = ENVIRON("REMOTE_HOST")    '  host.
CurUser = CurHost

LoadUserBase()

OPEN CONS FOR INPUT AS #4
LogLine("Received connection from " + CurUser + ".")
IF WriteAccess = 1 THEN
    PRINT "200 " + Banner + CRLF; ' Write access (potentially) ok
  ELSE
    PRINT "201 " + Banner + CRLF; ' Write access not ok (read-only server)
END IF
DO
  LineBuff = LineInput(4)
  ExplodeCommand(LineBuff)
  LogLine("[client tells] " + LineBuff)
  IF LastCommand(1) = "AUTHINFO" THEN
      AskForAuth(LastCommand(2), LastCommand(3))
    ELSE
      IF GlobalAuth = 1 AND LEN(AuthUser) = 0 THEN
          RequireAuth()
        ELSE
          SELECT CASE UCASE(LastCommand(1))
            CASE "LIST"
              ListGroups(LastCommand(2), LastCommand(3))
            CASE "LISTGROUP"
              SelectListGroup(LastCommand(2), LastCommand(3))
            CASE "ARTICLE"   ' nnn or <msgid>
              ReturnArticle(LastCommand(2), "ARTICLE")
            CASE "HEAD"   ' nnn or <msgid>
              ReturnArticle(LastCommand(2), "HEAD")
            CASE "BODY"   ' nnn or <msgid>
              ReturnArticle(LastCommand(2), "BODY")
            CASE "STAT"   ' nnn or <msgid>
              ReturnArticle(LastCommand(2), "STAT")
            CASE "NEXT"
              NextLast("NEXT")
            CASE "LAST"
              NextLast("LAST")
            CASE "MODE"
              SwitchMode(LastCommand(2))
            CASE "NEWGROUPS"
              NewGroups(LastCommand(2), LastCommand(3), LastCommand(4))
            CASE "POST"
              PostingProcess()
            CASE "GROUP"
              SelectGroup(LCASE(LastCommand(2)))
            CASE "OVER"
              Over(LastCommand(2))
            CASE "XOVER"
              Xover(LastCommand(2))
            CASE "XGTITLE"
              XgTitle(LastCommand(2))
            CASE "HELP"
              Help()
            CASE "DATE"
              PRINT ReturnRFC2980date() + CRLF;
              LogLine("Current GMT date returned.")
            CASE "QUIT"
              PRINT "205 See you later!" + CRLF;
              ConnClose()
            CASE "CAPABILITIES"
              AnnounceCapabilities()
            CASE ELSE
              PRINT "500 I don't know such command." + CRLF;
              LogLine("Unrecognized command.")
          END SELECT
      END IF
  END IF
LOOP
CLOSE #4
