REM
REM    This file is part of the BasicNews project. It contains all
REM    sub-routines and functions called by the BasicNews engine.
REM

REM  ***  Here comes all required includes  ***
#INCLUDE ONCE "crt/linux/unistd.bi"  ' Required for GetPID()
#INCLUDE ONCE "vbcompat.bi"

RANDOMIZE TIMER   ' Feed the random routines with timer's value


REM  ***  There come all sub-routines and functions  ***


FUNCTION IsRangeValid(SpecRange AS STRING) AS BYTE
  REM  Checks wheter a range (as in RFC 3977) is valid or not.
  REM  Returns 1 if the range is okay, 0 otherwise.
  DIM AS BYTE Result = 0
  DIM AS STRING WorkRange, TempString1, TempString2
  DIM AS INTEGER DashPosition
  WorkRange = SpecRange
  DashPosition = INSTR(WorkRange, "-")
  SELECT CASE DashPosition
    CASE 0
      IF "" & VAL(WorkRange) = SpecRange THEN Result = 1
    CASE 1
      Result = 0
    CASE LEN(WorkRange)
      WorkRange = LEFT(WorkRange, LEN(WorkRange) - 1)
      IF "" & VAL(WorkRange) & "-" = SpecRange THEN Result = 1
    CASE ELSE
      TempString1 = MID(WorkRange, 1, DashPosition - 1)
      TempString2 = MID(WorkRange, DashPosition + 1)
      IF "" & VAL(TempString1) = TempString1 AND "" & VAL(TempString2) = TempString2 THEN Result = 1
  END SELECT
  RETURN Result
END FUNCTION


FUNCTION RangeLow(SpecRange AS STRING) AS INTEGER
  REM Take a range in the format "numberlow-numberhigh", and returns its low value.
  REM Possible ranges: "2-78" ; "4" ; "56-"
  DIM AS INTEGER FirstInRange, LastInRange, x
  x = INSTR(SpecRange, "-")
  IF x = 0 THEN
      FirstInRange = VAL(SpecRange)
      LastInRange = VAL(SpecRange)
    ELSE
      FirstInRange = VAL(MID(SpecRange, 1, x - 1))
      IF x = LEN(SpecRange) THEN
          LastInRange = -1
        ELSE
          LastInRange = VAL(MID(SpecRange, x + 1))
      END IF
  END IF
  RETURN FirstInRange
END FUNCTION


FUNCTION RangeHigh(SpecRange AS STRING) AS INTEGER
  REM Take a range in the format "numberlow-numberhigh", and returns its high value (or -1 if no high value found).
  REM Possible ranges: "2-78" ; "4" ; "56-"
  DIM AS INTEGER FirstInRange, LastInRange, x
  x = INSTR(SpecRange, "-")
  IF x = 0 THEN
      FirstInRange = VAL(SpecRange)
      LastInRange = VAL(SpecRange)
    ELSE
      FirstInRange = VAL(MID(SpecRange, 1, x - 1))
      IF x = LEN(SpecRange) THEN
          LastInRange = -1
        ELSE
          LastInRange = VAL(MID(SpecRange, x + 1))
      END IF
  END IF
  RETURN LastInRange
END FUNCTION


FUNCTION NowGMT() AS DOUBLE
  DIM AS DOUBLE Wynik
  DIM AS INTEGER res, YearGMT, MonthGMT, DayGMT, HourGMT, MinuteGMT, SecondGMT
  DIM AS STRING timestr
  DIM AS time_t rawtime
  DIM AS tm Ptr timeptr
  time_(@rawtime)
  timeptr=gmtime(@rawtime)
  timestr=Space(128)
  res = strftime(Strptr(timestr),128,"%Y",timeptr)
  YearGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%m",timeptr)
  MonthGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%d",timeptr)
  DayGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%H",timeptr)
  HourGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%M",timeptr)
  MinuteGMT = VAL(LEFT(timestr, res))
  res = strftime(Strptr(timestr),128,"%S",timeptr)
  SecondGMT = VAL(LEFT(timestr, res))
  Wynik = DATESERIAL(YearGMT, MonthGMT, DayGMT) + TIMESERIAL(HourGMT, MinuteGMT, SecondGMT)
  RETURN Wynik
END FUNCTION


FUNCTION Like(BYVAL main AS ZSTRING PTR, BYVAL pat AS ZSTRING PTR) AS INTEGER
  REM  The "Like" function is used by the "DoesWildmatMatch" one.
  REM  "Like" check a given string against a given wildcard, and tells
  REM  wheter they match or not.
  DIM AS UBYTE a, b, d
  DIM AS UBYTE PTR r, p
  DIM AS INTEGER star
  DO
    d = *pat
    pat += 1
    IF d = 0 THEN
        RETURN (star <> 0) OR (*cptr(UBYTE PTR,main) = 0)
      ELSEIF d = ASC("*") THEN
        star = 1
      ELSEIF d = ASC("?") THEN
        IF *cptr(UBYTE PTR, main) = 0 THEN RETURN 0
        main += 1
      ELSE
        EXIT DO
    END IF
  LOOP
  b = d
  DO
    a = *main
    main += 1
    IF a = b THEN
      r = main
      p = pat
      DO
        d = *p
        p += 1
        IF d = ASC("*") THEN
            IF like(r,p-1) = -1 THEN RETURN -1
            EXIT DO
          ELSEIF d = 0 THEN
            IF *r = 0 THEN RETURN -1
            EXIT DO
        END IF
        r += 1
      LOOP WHILE (*(r-1) = d) OR (d = ASC("?"))
    END IF
  LOOP WHILE star <> 0 AND a <> 0
  RETURN 0
END FUNCTION


FUNCTION DoesWildmatMatch(Wildmat AS STRING, WorkString AS STRING) AS BYTE
  DIM AS INTEGER x, WildmatsCount
  DIM AS BYTE Result = 0
  REDIM Wildmats(1 TO 1) AS STRING
  REM First, let's explode all possible wildmats
  IF LEN(Wildmat) > 0 THEN WildmatsCount = 1
  FOR x = 1 TO LEN(Wildmat)
    IF MID(Wildmat, x, 1) = "," THEN
        IF UBOUND(Wildmats) = WildmatsCount THEN REDIM PRESERVE Wildmats(1 TO WildmatsCount * 2) AS STRING
        WildmatsCount += 1
      ELSE
        Wildmats(WildmatsCount) += MID(Wildmat, x, 1)
    END IF
  NEXT x
  REM Now, we'll process each wildmat found through the wildcard procedure
  FOR x = 1 TO WildmatsCount
    IF Like(WorkString, Wildmats(x)) <> 0 THEN Result = 1
  NEXT x
  RETURN Result
END FUNCTION


FUNCTION GetFileSize(Filepath AS STRING) AS INTEGER
  REM  Returns the size (in bytes) of a given file, or -1 if the
  REM  file could not be opened.
  DIM AS INTEGER Result, FileHandler, OpenResult
  FileHandler = FREEFILE
  OpenResult = OPEN(Filepath, FOR BINARY, AS #FileHandler)
  IF OpenResult = 0 THEN
      Result = LOF(FileHandler)
    ELSE
      Result = -1
  END IF
  CLOSE #FileHandler
  RETURN Result
END FUNCTION


SUB LogLine(TextToLog AS STRING = "")
  DIM AS INTEGER FileSocket
  DIM AS STRING TimeStamp
  DIM AS DOUBLE LogNOW
  IF VerboseMode > 0 THEN
    LogNOW = NOW
    TimeStamp = CHR(34) + FORMAT(YEAR(LogNOW), "0000") + "-" + FORMAT(MONTH(LogNOW), "00") + "-" + FORMAT(DAY(LogNOW), "00") + " " + FORMAT(HOUR(LogNOW), "00") + ":" + FORMAT(MINUTE(LogNOW), "00") + ":" + FORMAT(SECOND(LogNOW), "00") + CHR(34)
    FileSocket = FREEFILE
    OPEN LogFile FOR APPEND AS #FileSocket
      PRINT #FileSocket, "PID="""; FORMAT(GetPID(), "00000"); """ Date="; TimeStamp; " User="""; CurUser; """ Msg="""; TextToLog; """"
    CLOSE #FileSocket
  END IF
END SUB


SUB Help()
  PRINT "100 Help yourself, buddy." + CRLF;
  PRINT " Check out following documents (BasicNews is working according to them):" + CRLF;
  PRINT "  * RFC 1036" + CRLF;
  PRINT "  * RFC 2980" + CRLF;
  PRINT "  * RFC 3977" + CRLF;
  PRINT "  * RFC 4643" + CRLF;
  PRINT "." + CRLF;
  LogLine("Help displayed.")
END SUB


FUNCTION GetGroupCreationTime(NewsGroupName AS STRING) AS STRING
  DIM AS STRING CreationFilename, Result, WorkLine
  DIM AS INTEGER FileHandler, GroupYear, GroupMonth, GroupDay, GroupHour, GroupMinute, GroupSecond
  DIM AS DOUBLE CurGMT
  DIM AS BYTE ResetFile = 0
  CurGMT = NowGMT()
  CreationFilename = SpoolPath + NewsGroupName + "/creation.time"
  FileHandler = FREEFILE
  IF FileExists(CreationFilename) THEN
      OPEN CreationFilename FOR INPUT AS #FileHandler
      LINE INPUT #FileHandler, WorkLine
      IF LEN(WorkLine) = 19 THEN
          GroupYear = VAL(MID(WorkLine, 1, 4))
          GroupMonth = VAL(MID(WorkLine, 5, 2))
          GroupDay = VAL(MID(WorkLine, 7, 2))
          GroupHour = VAL(MID(WorkLine, 10, 2))
          GroupMinute = VAL(MID(WorkLine, 12, 2))
          GroupSecond = VAL(MID(WorkLine, 14, 2))
          Result = FORMAT(GroupYear, "0000") + FORMAT(GroupMonth, "00") + FORMAT(GroupDay, "00") + " " + FORMAT(GroupHour, "00") + FORMAT(GroupMinute, "00") + FORMAT(GroupSecond, "00") + " GMT"
        ELSE
          ResetFile = 1
      END IF
      CLOSE #FileHandler
    ELSE
      ResetFile = 1
  END IF
  IF ResetFile = 1 THEN
      LogLine("New local group found: " + NewsGroupName + ". Creation time recorded.")
      Result = FORMAT(YEAR(CurGMT), "0000") + FORMAT(MONTH(CurGMT), "00") + FORMAT(DAY(CurGMT), "00") + " " + FORMAT(HOUR(CurGMT), "00") + FORMAT(MINUTE(CurGMT), "00") + FORMAT(SECOND(CurGMT), "00") + " GMT"
      OPEN CreationFilename FOR OUTPUT AS #FileHandler
      PRINT #FileHandler, Result
      CLOSE #FileHandler
  END IF
  RETURN Result
END FUNCTION


SUB ConnClose()
  CLOSE #4
  LogLine("Connection closed, duration: " & FORMAT(TIMER - StartTime, "0.000") & " s.")
  END
END SUB


FUNCTION CheckTrimBOM(WorkString AS STRING) AS STRING  ' Remove any BOM UTF8 header from the line
  DIM Result AS STRING = WorkString
  DIM WorkStringLen AS INTEGER = LEN(WorkString)
  IF WorkStringLen >= 3 THEN
    IF MID(WorkString, 1, 3) = CHR(239) + CHR(187) + CHR(191) THEN
      IF WorkStringLen > 3 THEN
          Result = MID(WorkString, 4)
        ELSE
          Result = ""
      END IF
    END IF
  END IF
  RETURN Result
END FUNCTION


SUB About()
  PRINT "BasicNews v"; pVer; " Copyright (C) Mateusz Viste "; pDate
  PRINT
  PRINT "This program is free software: you can redistribute it and/or modify it under"
  PRINT "the terms of the GNU General Public License as published by the Free Software"
  PRINT "Foundation, either version 3 of the License, or (at your option) any later"
  PRINT "version."
  PRINT "This program is distributed in the hope that it will be useful, but WITHOUT ANY"
  PRINT "WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A"
  PRINT "PARTICULAR PURPOSE. See the GNU General Public License for more details."
  PRINT
  PRINT "BasicNews is a simple open-source news (NNTP) server for Linux. BasicNews is not a"
  PRINT "standalone daemon - it requires an inetd-compatible superserver to work."
  PRINT "I wrote BasicNews because I couldn't find any *simple* news server which would allow"
  PRINT "me to put a private newsgroup online in a short time. BasicNews is targetted to"
  PRINT "people who wish to run a small news server for personal purpose (private server, or"
  PRINT "a company's internal newsgroup). BasicNews is entirely written in FreeBASIC."
  PRINT
  PRINT "Homepage: "; homepage
  PRINT
  END
END SUB


FUNCTION GetMsgHeader(ArtFile AS STRING, Header AS STRING) AS STRING
  DIM AS INTEGER ArtHandler, SeparatorPos, LineCounter = 0
  DIM AS BYTE HeaderEnded = 0
  DIM AS STRING TempString, Token, Value, Result
  ArtHandler = FREEFILE
  OPEN ArtFile FOR INPUT AS #ArtHandler
  WHILE NOT EOF(ArtHandler) AND UCASE(Token) <> UCASE(Header) AND LineCounter < 512 AND HeaderEnded = 0
    Token = ""
    Value = ""
    LINE INPUT #ArtHandler, TempString
    IF TempString = "" THEN HeaderEnded = 1
    LineCounter += 1
    SeparatorPos = INSTR(TempString, ":")
    IF SeparatorPos > 1 THEN
      Token = LEFT(TempString, SeparatorPos - 1)
      Value = TRIM(MID(TempString, SeparatorPos + 1))
    END IF
  WEND
  CLOSE #ArtHandler
  IF UCASE(Token) = UCASE(Header) THEN Result = Value ELSE Result = ""
  RETURN Result
END FUNCTION


SUB RegenerateOverview(Newsgroup AS STRING)
  DIM XoverField(1 TO 9) AS STRING ' Fmt: ArtNum|Subject|Author|Date|MsgID|Ref|bytes|lines|Xref
  DIM AS STRING FileName
  DIM AS UINTEGER FileAttrib
  DIM AS INTEGER TmpFileHandler

  TmpFileHandler = FREEFILE
  OPEN SpoolPath & Newsgroup & "/overview.idx" FOR OUTPUT AS #TmpFileHandler
  FileName = DIR(SpoolPath + Newsgroup + "/*", &h37, FileAttrib)
  WHILE FileName <> ""
    IF LEFT(FileName, 1) <> "." THEN   ' Can't start by a dot.
      IF NOT FileAttrib AND &h10 THEN  ' Can't be a directory.
        IF VAL(FileName) > 0 THEN   ' Filename must be a positive number.
          XoverField(1) = "" & VAL(FileName)
          XoverField(2) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "Subject")
          XoverField(3) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "From")
          XoverField(4) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "Date")
          XoverField(5) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "Message-ID")
          XoverField(6) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "References")
          XoverField(7) = "" & GetFileSize(SpoolPath + Newsgroup + "/" + FileName)
          XoverField(8) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "Lines")
          XoverField(9) = GetMsgHeader(SpoolPath + Newsgroup + "/" + FileName, "Xref")
          PRINT #TmpFileHandler, XoverField(1) + CHR(9) + XoverField(2) + CHR(9) + XoverField(3) + CHR(9) + XoverField(4) + CHR(9) + XoverField(5) + CHR(9) + XoverField(6) + CHR(9) + XoverField(7) + CHR(9) + XoverField(8) + CHR(9) + "Xref: " + XoverField(9) + CRLF;
        END IF
      END IF
    END IF
    FileName = DIR(FileAttrib)
  WEND
  CLOSE #TmpFileHandler
END SUB


SUB CheckOverviewFile(GroupToCheck AS STRING)
  DIM AS INTEGER OpenResult, TmpFileHandler
  TmpFileHandler = FREEFILE
  OpenResult = OPEN(SpoolPath & GroupToCheck & "/overview.idx", FOR INPUT, AS #TmpFileHandler)
  IF OpenResult = 0 THEN
      CLOSE #TmpFileHandler
    ELSE
      LogLine("No overview file found for " + GroupToCheck + ". A new overview file has been generated.")
      RegenerateOverview(GroupToCheck)
  END IF
END SUB


SUB LoadGroupList() ' May be called several times without harm (eg. to refresh list of groups)
  DIM AS STRING FileName
  DIM AS UINTEGER FileAttrib
  GroupCount = 0
  FileName = DIR(SpoolPath + "*", &h37, FileAttrib)
  WHILE FileName <> ""
    IF LEFT(FileName, 1) <> "." AND (FileAttrib AND &h10) THEN
      IF GroupCount = UBOUND(GroupList) THEN REDIM PRESERVE GroupList(1 TO GroupCount * 2) AS STRING
      GroupCount += 1
      GroupList(GroupCount) = FileName
    END IF
    FileName = DIR(FileAttrib)
  WEND
END SUB


FUNCTION SearchMsgFileByID(IdToken AS STRING) AS STRING
  DIM AS INTEGER x, y, z, t, OverviewFileHandler
  DIM AS STRING FileFound, LineBuff
  DIM TempField(1 TO 6) AS STRING
  LoadGroupList()
  OverviewFileHandler = FREEFILE
  x = 0
  WHILE x < GroupCount AND LEN(FileFound) = 0
    x += 1
    CheckOverviewFile(GroupList(x))
    OPEN SpoolPath + GroupList(x) + "/overview.idx" FOR INPUT AS #OverviewFileHandler
    WHILE NOT EOF(OverviewFileHandler) AND LEN(FileFound) = 0
      LINE INPUT #OverviewFileHandler, LineBuff
      z = 0 : y = 1
      FOR t = 1 TO UBOUND(TempField)
        TempField(t) = ""    ' Clearing all fields
      NEXT t
      WHILE y < 6 AND z < LEN(LineBuff)
        z += 1
        IF LineBuff[z - 1] <> 9 THEN
            TempField(y) += MID(LineBuff, z, 1)
          ELSE
            y += 1
        END IF
      WEND
      IF TempField(5) = IdToken THEN FileFound = SpoolPath + GroupList(x) + "/" + TempField(1)
    WEND
    CLOSE #OverviewFileHandler
  WEND
  RETURN FileFound
END FUNCTION


FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
 STATIC CfgTable(1 TO 2, 0 TO 255) AS STRING
 DIM AS STRING CfgReturnString, CfgTmpBuffer
 DIM AS INTEGER Counter
 IF CfgTable(1, 0) <> CFGfile THEN
   DIM AS INTEGER CfgFileHandler, CfgColonPos, Counter
   CfgFileHandler = FREEFILE
   CfgReturnString = ""
   Counter = 0
   CfgTable(1, 0) = CFGfile
   IF DIR(CFGfile) <> "" THEN
      OPEN CFGfile FOR INPUT AS #CfgFileHandler
      DO
         Counter += 1
         LINE INPUT #CfgFileHandler, CfgTmpBuffer
         CfgTmpBuffer = CheckTrimBOM(CfgTmpBuffer)
         IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
           CfgColonPos = INSTR(CfgTmpBuffer, "=")
           CfgTable(1, Counter) = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
           CfgTable(2, Counter) = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
         END IF
      LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
      CLOSE #CfgFileHandler
   END IF
   CfgTable(2, 0) = STR(Counter)
 END IF
 Counter = 0
 DO
   Counter += 1
   IF UCASE(CfgTable(1, Counter)) = UCASE(CFGField) THEN CfgReturnString = CfgTable(2, Counter)
 LOOP UNTIL CfgReturnString <> "" OR Counter >= VAL(CfgTable(2, 0))
 RETURN CfgReturnString
END FUNCTION


FUNCTION LineInput(SocketToListenTo AS INTEGER = 0) AS STRING
  DIM AS STRING Wynik, CharBuff
  DIM AS DOUBLE LastData
  DIM AS BYTE TimeoutAbort = 0
  LastData = TIMER
  DO
    CharBuff = INPUT(1, SocketToListenTo)
    IF CharBuff <> CHR(13) AND CharBuff <> CHR(10) AND CharBuff <> "" THEN Wynik += CharBuff: LastData = TIMER
    IF TIMER - LastData > 180 THEN TimeoutAbort = 1
    IF TIMER - LastData > 0.1 THEN SLEEP 1,1     ' Do not hog the CPU when waiting for input...
  LOOP UNTIL TimeoutAbort = 1 OR LEN(Wynik) > 1024 OR CharBuff = CHR(10)
  IF LEN(Wynik) > 1024 THEN
    LogLine("I/O OVERFLOW - Treatement aborted. Dataline too long (>1 KiB).")
    SLEEP 500, 1
    ConnClose()
  END IF
  IF TimeoutAbort = 1 THEN
    REM  When the timer expires, the server SHOULD close the connection
    REM  without sending any response to the client. [RFC 3977]
    LogLine("I/O TIMEOUT - Treatement aborted. Waiting period too long (>180s).")
    ConnClose()
  END IF
  RETURN Wynik
END FUNCTION


SUB ExplodeCommand(CommandString AS STRING)
  DIM AS INTEGER x, y
  FOR x = 1 TO UBOUND(LastCommand)
    LastCommand(x) = ""     ' Clear last command
  NEXT x
  x = 0
  y = 1
  DO
    IF CommandString[x] <> 32 THEN
        LastCommand(y) += CHR(CommandString[x])
      ELSE
        y += 1
    END IF
    x += 1
  LOOP UNTIL x > LEN(CommandString) - 1 OR y > UBOUND(LastCommand)
END SUB


SUB RequireAuth()
  PRINT "480 Permission denied." + CRLF;
  LogLine("Request rejected - authentication required...")
END SUB


SUB PrintFile(FileToPrint AS STRING, SpecificRequest AS STRING = "") ' Print a file on the output...
  REM  SpecificRequest may be "HEAD" or "BODY"
  DIM AS STRING TempLine
  DIM AS INTEGER FileHandler, x
  DIM AS BYTE PrintHeader, PrintBody, HeaderDone, FinalDot
  SELECT CASE UCASE(SpecificRequest)
    CASE "BODY"
      PrintHeader = 0
      PrintBody = 1
      LogLine("Return the BODY part of the " + FileToPrint + " file.")
    CASE "HEAD"
      PrintHeader = 1
      PrintBody = 0
      LogLine("Return the HEAD part of the " + FileToPrint + " file.")
    CASE ELSE
      PrintHeader = 1
      PrintBody = 1
      LogLine("Return the " + FileToPrint + " file.")
  END SELECT
  FileHandler = FREEFILE
  OPEN FileToPrint FOR INPUT AS #FileHandler
  WHILE NOT EOF(FileHandler) AND (HeaderDone = 0 OR PrintBody = 1) AND FinalDot = 0
    LINE INPUT #FileHandler, TempLine
    IF HeaderDone = -1 THEN HeaderDone = 1
    IF LEN(TempLine) = 0 AND HeaderDone = 0 THEN HeaderDone = -1
    IF HeaderDone = 0 AND PrintHeader = 1 THEN PRINT TempLine + CRLF;
    IF HeaderDone = 1 AND PrintBody = 1 THEN PRINT TempLine + CRLF;
    IF HeaderDone = -1 AND PrintHeader = 1 AND PrintBody = 1 THEN PRINT TempLine + CRLF;
    IF TempLine = "." THEN FinalDot = 1
  WEND
  IF FinalDot = 0 THEN PRINT "." + CRLF;  ' Print the terminator (dot), if not provided yet.
  CLOSE #FileHandler
END SUB


FUNCTION GetGroupProperties(NewsGroupName AS STRING) AS NewsGroup
  DIM Result AS NewsGroup
  Result.Name = NewsGroupName
  Result.Description = ""
  Result.Auth = 0
  Result.FirstArticle = 0
  Result.LastArticle = 0
  Result.ArticlesCount = 0
  Result.WriteAccess = 0
  Result.CurArticle = 0
  Result.MaxArticleSize = 0
  DIM AS INTEGER DirIndex
  DIM AS UINTEGER FileAttrib
  DIM AS STRING FileName
  FileName = DIR(SpoolPath + NewsGroupName + "/*", &h37, FileAttrib)
    DirIndex = 0
    WHILE FileName <> ""
      IF LEFT(FileName, 1) <> "." THEN   ' Can't start by a dot.
        IF NOT FileAttrib AND &h10 THEN  ' Can't be a directory.
          IF VAL(FileName) > 0 THEN   ' Filename must be a positive number.
            DirIndex += 1
            IF VAL(FileName) < Result.FirstArticle OR Result.FirstArticle = 0 THEN Result.FirstArticle = VAL(FileName)
            IF VAL(FileName) > Result.LastArticle THEN Result.LastArticle = VAL(FileName)
          END IF
        END IF
      END IF
      FileName = DIR(FileAttrib)
    WEND
    Result.CreationTime = GetGroupCreationTime(NewsGroupName)
    Result.CurArticle = CurGroup.FirstArticle
    Result.ArticlesCount = DirIndex
    Result.Description = ReadCFG(SpoolPath + Result.Name + ".conf", "Description")
    Result.MaxArticleSize = VAL(ReadCFG(SpoolPath + Result.Name + ".conf", "MaxArticleSize"))
    Result.Auth = VAL(ReadCFG(SpoolPath + Result.Name + ".conf", "Auth"))
    IF WriteAccess = 1 THEN
        Result.WriteAccess = VAL(ReadCFG(SpoolPath + Result.Name + ".conf", "WriteAccess"))
      ELSE
        Result.WriteAccess = 0
    END IF
    RETURN Result
END FUNCTION


FUNCTION CheckAuth(CheckGroup AS STRING) AS BYTE
  REM  Returns 1 if user authentication is good enough
  REM  for the given group, returns 0 otherwise.
  DIM AS BYTE Result
  DIM AS STRING UserBuff
  DIM AS INTEGER AuthFileHandler, OpenResult
  DIM TempGroup AS NewsGroup
  TempGroup = GetGroupProperties(CheckGroup)
  IF TempGroup.Auth = 0 THEN
      Result = 1
    ELSE
      IF LEN(AuthUser) = 0 THEN
          Result = 0
        ELSE
          Result = 0
          AuthFileHandler = FREEFILE
          OpenResult = OPEN(SpoolPath & CheckGroup & ".users", FOR INPUT, AS #AuthFileHandler)
          IF OpenResult = 0 THEN
          WHILE NOT EOF(AuthFileHandler) AND Result = 0
              LINE INPUT #AuthFileHandler, UserBuff
              UserBuff = TRIM(UserBuff)
              IF LEN(UserBuff) > 0 THEN
                IF LEFT(UserBuff, 1) <> "#" AND UserBuff = AuthUser THEN Result = 1
              END IF
            WEND
            CLOSE #AuthFileHandler
          END IF
      END IF
  END IF
  RETURN Result
END FUNCTION


FUNCTION DoesGroupExists(GroupToCheck AS STRING) AS BYTE
  DIM AS BYTE Result = 0
  DIM AS STRING TempString
  DIM AS UINTEGER GetAttribs
  IF LEN(GroupToCheck) > 0 THEN
    TempString = DIR(SpoolPath + GroupToCheck, &h37, GetAttribs)
    IF GetAttribs AND &h10 THEN Result = 1
  END IF
  RETURN Result
END FUNCTION


SUB SelectGroup(GroupName AS STRING)
  IF LEN(GroupName) > 0 THEN
      IF DoesGroupExists(GroupName) = 1 THEN
          IF CheckAuth(GroupName) = 0 THEN
              RequireAuth()
            ELSE
              CurGroup = GetGroupProperties(GroupName)
              CurGroup.CurArticle = CurGroup.FirstArticle   ' Set the current article to the first one (mandatory for the GROUP command)
              PRINT "211 " & CurGroup.ArticlesCount & " " & CurGroup.FirstArticle & " " & CurGroup.LastArticle & " " & GroupName & " group selected." + CRLF;
              LogLine("Group " + CurGroup.Name + " selected.")
          END IF
        ELSE
          PRINT "411 No such news group here!" + CRLF;
          LogLine("Group not found (" + GroupName + ")")
      END IF
    ELSE
      PRINT "501 You have to specify a group name, you dumb ass!" + CRLF;
      LogLine("No group specified!")
  END IF
END SUB


SUB SelectListGroup(GroupName AS STRING, ArtRange AS STRING)
  DIM AS INTEGER FirstArt, LastArt, x
  IF LEN(GroupName) = 0 THEN GroupName = CurGroup.Name
  IF LEN(GroupName) > 0 THEN
      IF DoesGroupExists(GroupName) = 1 THEN
          IF CheckAuth(GroupName) = 0 THEN
              RequireAuth()
            ELSE
              CurGroup = GetGroupProperties(GroupName)
              CurGroup.CurArticle = CurGroup.FirstArticle   ' Set the current article to the first one (mandatory for the GROUP command)
              PRINT "211 " & CurGroup.ArticlesCount & " " & CurGroup.FirstArticle & " " & CurGroup.LastArticle & " " & GroupName & " group selected." + CRLF;
              IF LEN(ArtRange) > 0 THEN
                  FirstArt = RangeLow(ArtRange)
                  LastArt = RangeHigh(ArtRange)
                  IF LastArt = -1 THEN LastArt = CurGroup.LastArticle
                ELSE
                  FirstArt = CurGroup.FirstArticle
                  LastArt = CurGroup.LastArticle
              END IF
              IF FirstArt < CurGroup.FirstArticle THEN FirstArt = CurGroup.FirstArticle
              IF LastArt > CurGroup.LastArticle THEN LastArt = CurGroup.LastArticle
              REM Here we'll display all articles.
              FOR x = FirstArt TO LastArt
                IF FileExists(SpoolPath + CurGroup.Name + "/" + STR(x)) <> 0 THEN PRINT "" & x & CRLF;
              NEXT x
              PRINT "." + CRLF;
              LogLine("Group " + CurGroup.Name + " selected and articles listed.")
          END IF
        ELSE
          PRINT "411 No such news group here!" + CRLF;
          LogLine("Group not found (" + GroupName + ")")
      END IF
    ELSE
      PRINT "412 No currently selected group." + CRLF;
      LogLine("No group specified, and no group currently selected!")
  END IF
END SUB


FUNCTION ReturnRFC2980date() AS STRING
  DIM AS STRING Result
  DIM AS DOUBLE RightNow
  RightNow = NowGMT()
  Result = "111 " + FORMAT(YEAR(RightNow), "0000") + FORMAT(MONTH(RightNow), "00") + FORMAT(DAY(RightNow), "00") + FORMAT(HOUR(RightNow), "00") + FORMAT(MINUTE(RightNow), "00") + FORMAT(SECOND(RightNow), "00")
  RETURN Result
END FUNCTION


SUB ListGroups(SecParameter AS STRING, ThirdParameter AS STRING)
  DIM AS STRING TempLine
  DIM AS NewsGroup TempGroup
  DIM AS INTEGER x
  LoadGroupList()   ' Let's retrieve the list of local groups...
  SELECT CASE UCASE(SecParameter)
    CASE "", "ACTIVE"
      PRINT "215 List of available newsgroups follows" + CRLF;
      FOR x = 1 TO GroupCount
        TempGroup = GetGroupProperties(GroupList(x))
        TempLine = TempGroup.Name & " " & TempGroup.LastArticle & " " & TempGroup.FirstArticle & " "
        IF TempGroup.WriteAccess = 1 THEN TempLine += "y" ELSE TempLine += "n"
        IF LEN(ThirdParameter) > 0 THEN
            IF DoesWildmatMatch(LCASE(ThirdParameter), LCASE(TempGroup.Name)) <> 0 THEN PRINT TempLine + CRLF;
          ELSE
            PRINT TempLine + CRLF;
        END IF
      NEXT x
      PRINT "." + CRLF;
      LogLine("List of available newsgroups")
    CASE "NEWSGROUPS"
      PRINT "215 List of available newsgroups follows" + CRLF;
      FOR x = 1 TO GroupCount
        TempGroup = GetGroupProperties(GroupList(x))
        TempLine = TempGroup.Name & CHR(9) & TempGroup.Description
        IF LEN(ThirdParameter) > 0 THEN
            IF DoesWildmatMatch(LCASE(ThirdParameter), LCASE(TempGroup.Name)) <> 0 THEN PRINT TempLine + CRLF;
          ELSE
            PRINT TempLine + CRLF;
        END IF
      NEXT x
      PRINT "." + CRLF;
      LogLine("List of available newsgroups")
    CASE "OVERVIEW.FMT" ' Change the informations below if you modify the XOVER/OVER command!
      PRINT "215 Order of fields in overview database:" + CRLF;
      PRINT "Subject:" + CRLF;
      PRINT "From:" + CRLF;
      PRINT "Date:" + CRLF;
      PRINT "Message-ID:" + CRLF;
      PRINT "References:" + CRLF;
      PRINT ":bytes" + CRLF;
      PRINT ":lines" + CRLF;
      PRINT "Xref:full" + CRLF;
      PRINT "." + CRLF;
      LogLine("Sending the overview format.")
    CASE ELSE
      PRINT "501 I do not support LIST with this parameter." + CRLF;
      LogLine("Unsupported parameter.")
  END SELECT
END SUB


SUB NewGroups(DateFrom AS STRING, TimeFrom AS STRING, TimeZone AS STRING)
  DIM AS INTEGER x, DayWanted, MonthWanted, YearWanted, HourWanted, MinuteWanted, SecondWanted, NewGroupsCount = 0, CurCentury, CurYear
  DIM AS NewsGroup TempGroup
  DIM AS DOUBLE DateGroup, DateWanted
  DIM AS BYTE BrokenClient = 0
  DIM AS STRING TempLine
  LoadGroupList()    ' Retrieve the most up-to-date list of local groups...
  SELECT CASE LEN(DateFrom)
    CASE 6
      YearWanted = VAL(MID(DateFrom, 1, 2))
      CurYear = YEAR(NowGMT())
      CurCentury =  CurYear \ 100  ' What's the current
      CurCentury *= 100            ' century?
      IF YearWanted + CurCentury > CurYear THEN
          YearWanted += (CurCentury - 1)
        ELSE
          YearWanted += CurCentury
      END IF
      MonthWanted = VAL(MID(DateFrom, 3, 2))
      DayWanted = VAL(MID(DateFrom, 5, 2))
    CASE 8
      YearWanted = VAL(MID(DateFrom, 1, 4))
      MonthWanted = VAL(MID(DateFrom, 5, 2))
      DayWanted = VAL(MID(DateFrom, 7, 2))
    CASE ELSE
      BrokenClient = 1
  END SELECT
  SELECT CASE LEN(TimeFrom)
    CASE 6
      HourWanted = VAL(MID(TimeFrom, 1, 2))
      MinuteWanted = VAL(MID(TimeFrom, 3, 2))
      SecondWanted = VAL(MID(TimeFrom, 5, 2))
    CASE ELSE
      BrokenClient = 1
  END SELECT
  IF LEN(TimeZone) > 0 AND UCASE(TimeZone) <> "GMT" THEN BrokenClient = 1
  IF BrokenClient = 0 THEN
      PRINT "231 List of new groups follows." + CRLF;
      DateWanted = DATESERIAL(YearWanted, MonthWanted, DayWanted) + TIMEVALUE(FORMAT(HourWanted, "00") + ":" + FORMAT(MinuteWanted, "00") + ":" + FORMAT(SecondWanted, "00"))
      FOR x = 1 TO GroupCount
        TempGroup = GetGroupProperties(GroupList(x))
        DateGroup = DATESERIAL(VAL(MID(TempGroup.CreationTime, 1, 4)), VAL(MID(TempGroup.CreationTime, 5, 2)), VAL(MID(TempGroup.CreationTime, 7, 2))) + TIMEVALUE(MID(TempGroup.CreationTime, 10, 2) + ":" + MID(TempGroup.CreationTime, 12, 2) + ":" + MID(TempGroup.CreationTime, 14, 2))
        IF DateGroup > DateWanted THEN
          NewGroupsCount += 1
          TempLine = TempGroup.Name & " " & TempGroup.LastArticle & " " & TempGroup.FirstArticle & " "
          IF TempGroup.WriteAccess = 1 THEN TempLine += "y" ELSE TempLine += "n"
          PRINT TempLine + CRLF;
        END IF
      NEXT x
      PRINT "." + CRLF;
      LogLine("List of new groups returned (" & NewGroupsCount & " groups).")
    ELSE
      PRINT "501 Invalid syntax. Should be yyyymmdd hhmmss [GMT]." + CRLF;
      LogLine("Syntax Error in provided date/time/timezone.")
  END IF
END SUB


SUB XgTitle(CmdParam AS STRING)
  DIM AS INTEGER x
  DIM AS NewsGroup TempGroup
  DIM AS STRING TempLine
  LoadGroupList()   ' Let's retrieve the list of local groups...
  PRINT "282 List of available newsgroups follows..." + CRLF;
  LogLine("List of available newsgroups.")
  FOR x = 1 TO GroupCount
    TempGroup = GetGroupProperties(GroupList(x))
    TempLine = TempGroup.Name & CHR(9) & TempGroup.Description
    IF LEN(CmdParam) > 0 THEN
        IF DoesWildmatMatch(LCASE(CmdParam), LCASE(TempGroup.Name)) <> 0 THEN PRINT TempLine + CRLF;
      ELSE
        PRINT TempLine + CRLF;
    END IF
  NEXT x
  PRINT "." + CRLF;
END SUB


SUB ReturnArticle(ArtID AS STRING, Request AS STRING) ' Request may be "ARTICLE", "STAT", "BODY" or "HEAD"
  DIM AS STRING TypeID, MsgID, TempString, TempString2
  DIM AS INTEGER ArtNum, FileHandle, OpenResult, AuthOk
  IF LEN(ArtID) = 0 THEN
      TypeID = "ArtNum"
      ArtNum = CurGroup.CurArticle
    ELSE
      IF LEFT(ArtID, 1) = "<" AND RIGHT(ArtID, 1) = ">" THEN
          TypeID = "MsgID"
        ELSE
          IF VAL(ArtID) > 0 THEN
              TypeID = "ArtNum"
              ArtNum = VAL(ArtID)
            ELSE
              TypeID = "INVALID"
          END IF
      END IF
  END IF
  REM Once we know the parameter's type, let's do the job!
  SELECT CASE TypeID
    CASE "ArtNum"
      IF LEN(CurGroup.Name) = 0 THEN
          PRINT "412 No newsgroup selected." + CRLF;
          LogLine("No newsgroup selected!")
        ELSE
          IF ArtNum = 0 THEN
              PRINT "420 Current article number is invalid." + CRLF;
              LogLine("Current article number is invalid.")
            ELSE
              FileHandle = FREEFILE
              OpenResult = OPEN(SpoolPath & CurGroup.Name & "/" & ArtNum, FOR INPUT, AS FileHandle)
              IF OpenResult = 0 THEN
                  CLOSE #FileHandle
                  MsgID = GetMsgHeader(SpoolPath & CurGroup.Name & "/" & ArtNum, "Message-ID")
                  CurGroup.CurArticle = ArtNum   ' Current Article pointer MUST be set when ARTICLE is called with a number
                  SELECT CASE UCASE(Request)
                    CASE "ARTICLE"
                      PRINT "220 " & ArtNum & " " & MsgID & " Article follows:" + CRLF;
                      PrintFile(SpoolPath & CurGroup.Name & "/" & ArtNum)
                    CASE "HEAD"
                      PRINT "221 " & ArtNum & " " & MsgID & " Article's header follows:" + CRLF;
                      PrintFile(SpoolPath & CurGroup.Name & "/" & ArtNum, "HEAD")
                    CASE "BODY"
                      PRINT "222 " & ArtNum & " " & MsgID & " Article's body follows:" + CRLF;
                      PrintFile(SpoolPath & CurGroup.Name & "/" & ArtNum, "BODY")
                    CASE "STAT"
                      PRINT "223 " & ArtNum & " " & MsgID & " The article exists and has been selected." + CRLF;
                      LogLine("The article exists and has been selected.")
                  END SELECT
                ELSE
                  PRINT "423 No article with that number." + CRLF;
                  LogLine("No article with that number.")
              END IF
          END IF
      END IF
    CASE "MsgID"  ' Remember: in this case the current article doesn't change!
      TempString = SearchMsgFileByID(ArtID)
      IF LEN(TempString) > 0 THEN
          REM Check authentication properties of all groups in which the message is present (and return it only if at least one of these groups is allowed in current state)
          TempString2 = GetMsgHeader(TempString, "Newsgroups")
          WHILE LEN(TempString2) > 0
            IF INSTR(TempString2, ",") > 0 THEN
                IF CheckAuth(MID(TempString2, 1, INSTR(TempString2, ",") - 1)) = 1 THEN AuthOk = 1
                TempString2 = MID(TempString2, INSTR(TempString2, ",") + 1)
              ELSE
                IF CheckAuth(TempString2) = 1 THEN AuthOk = 1
                TempString2 = ""
            END IF
          WEND
          TempString2 = GetMsgHeader(TempString, "Followup-to")
          WHILE LEN(TempString2) > 0
            IF INSTR(TempString2, ",") > 0 THEN
                IF CheckAuth(MID(TempString2, 1, INSTR(TempString2, ",") - 1)) = 1 THEN AuthOk = 1
                TempString2 = MID(TempString2, INSTR(TempString2, ",") + 1)
              ELSE
                IF CheckAuth(TempString2) = 1 THEN AuthOk = 1
                TempString2 = ""
            END IF
          WEND
          IF AuthOk = 0 THEN
              PRINT "480 Access denied to this article. Authentication required!" + CRLF;
              LogLine("Access denied (the article is in restricted groups)!")
            ELSE
              SELECT CASE UCASE(Request)
                CASE "ARTICLE"
                  PRINT "220 0 " + ArtID + " Article follows:" + CRLF; ' The returned article number must be 0!
                  PrintFile(TempString)
                CASE "HEAD"
                  PRINT "221 0 " + ArtID + " Article's header follows:" + CRLF;
                  PrintFile(TempString, "HEAD")
                CASE "BODY"
                  PRINT "222 0 " + ArtID + " Article's body follows:" + CRLF;
                  PrintFile(TempString, "BODY")
                CASE "STAT"
                  PRINT "223 0 " + ArtID + " The article exists and has been selected." + CRLF;
                  LogLine("The article exists and has been selected.")
            END SELECT
          END IF
        ELSE
          PRINT "430 No article with that Message-ID." + CRLF;
          LogLine("No article with that Message-ID.")
      END IF
    CASE ELSE
      PRINT "501 I don't understand your request." + CRLF;
      LogLine("Invalid request syntax!")
  END SELECT
END SUB


SUB NextLast(Request AS STRING)   ' Request may be "NEXT" or "LAST"
  DIM AS INTEGER FileHandle, ArtNum
  DIM AS STRING MsgID
  IF LEN(CurGroup.Name) = 0 THEN
      PRINT "412 No newsgroup selected." + CRLF;
      LogLine("No newsgroup selected!")
    ELSE
      ArtNum = CurGroup.CurArticle
      IF ArtNum = 0 THEN
          PRINT "420 Current article number is invalid." + CRLF;
          LogLine("Current article number is invalid!")
        ELSE
          DO
            SELECT CASE UCASE(Request)
              CASE "NEXT"
                ArtNum += 1
              CASE "LAST"
                ArtNum -= 1
            END SELECT
          LOOP UNTIL FileExists(SpoolPath & CurGroup.Name & "/" & ArtNum) <> 0 OR ArtNum > CurGroup.LastArticle OR ArtNum < CurGroup.FirstArticle
          IF ArtNum <= CurGroup.LastArticle AND ArtNum >= CurGroup.FirstArticle THEN
              MsgID = GetMsgHeader(SpoolPath & CurGroup.Name & "/" & ArtNum, "Message-ID")
              CurGroup.CurArticle = ArtNum
              PRINT "223 " & ArtNum & " " & MsgID & " Article found, and selected." + CRLF;
              LogLine("Article found, and selected.")
            ELSE
              SELECT CASE UCASE(Request)
                CASE "NEXT"
                  PRINT "421 No next article available." + CRLF;
                  LogLine("No next article available.")
                CASE "LAST"
                  PRINT "422 No previous article available." + CRLF;
                  LogLine("No previous article available.")
              END SELECT
          END IF
      END IF
  END IF
END SUB


SUB PostingProcess()
  REM  * There are some header the server should add: 
  REM   sender: AuthUser@HisIP
  REM   NNTP-Posting-Host: IpOfThePostingMachine
  REM   Lines: NumberOfLinesOfTheBody
  REM   Message-ID: MessageID (eg. <4123@ucbvax.Berkeley.ARPA>)
  REM   NNTP-Posting-Date: DateAtWhichTheServerReceivedThePost (Fri, 19-Nov-82 16:59:30 EST)
  REM  * Mandatory headers, provided by the sender: 
  REM   From, Subject, Newsgroups
  DIM XoverField(1 TO 9) AS STRING ' Fmt: ArtNum|Subject|Author|Date|MsgID|Ref|bytes|lines|Xref
  REDIM PostedContent(1 TO 32) AS STRING
  DIM AS INTEGER PostedLines = 0, BodyLines = 0, PostedSize = 0, FileHandler, x, y, LastHeaderLine = 0
  DIM AS BYTE Oversize = 0, HeaderDone = 0
  DIM AS DOUBLE RightNOW
  DIM AS STRING TempLine, MsgID, DateReceived, DstNewsGroups, FollowupTo, ErrString, ControlMsg, MsgPath
  DIM AS NewsGroup TempGroup
  DIM ShortWeekName(1 TO 7) AS STRING * 3 => {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"}
  DIM ShortMonth(1 TO 12) AS STRING * 3 => {"Jan", "Feb", "Mar", "Apr", "May", "Jun"_
                                            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"}
  IF WriteAccess = 1 THEN
      PRINT "340 I'm ready to receive your article. Send me your stuff now." + CRLF;
      LogLine("Receiving article...")
      DO
        TempLine = LineInput(4)
        IF PostedLines = UBOUND(PostedContent) THEN REDIM PRESERVE PostedContent(1 TO PostedLines * 2)
        PostedLines += 1
        IF HeaderDone = 1 THEN BodyLines += 1
        IF LEN(TempLine) = 0 THEN
          HeaderDone = 1
          IF LastHeaderLine = 0 THEN LastHeaderLine = PostedLines - 1
        END IF
        IF HeaderDone = 0 THEN  ' Detect some message's properties
          IF LEFT(UCASE(TempLine), 6) = "PATH: " THEN MsgPath = MID(TempLine, 7)
          IF LEFT(UCASE(TempLine), 12) = "NEWSGROUPS: " THEN DstNewsGroups = MID(TempLine, 13)
          IF LEFT(UCASE(TempLine), 13) = "FOLLOWUP-TO: " THEN FollowupTo = MID(TempLine, 14)
          IF LEFT(UCASE(TempLine), 9) = "CONTROL: " THEN ControlMsg = MID(TempLine, 10)
          IF LEFT(UCASE(TempLine), 9) = "SUBJECT: " THEN XoverField(2) = MID(TempLine, 10)
          IF LEFT(UCASE(TempLine), 6) = "FROM: " THEN XoverField(3) = MID(TempLine, 7)
          IF LEFT(UCASE(TempLine), 6) = "DATE: " THEN XoverField(4) = MID(TempLine, 7)
          IF LEFT(UCASE(TempLine), 12) = "REFERENCES: " THEN XoverField(6) = MID(TempLine, 13)
          'IF LEFT(UCASE(TempLine), 6) = "XREF: " THEN XoverField(9) = TempLine ' Xref has to be forged by the server!
        END IF
        PostedContent(PostedLines) = TempLine
        PostedSize += LEN(TempLine) + 2  ' Add two bytes for CR/LF chars
        IF CurGroup.MaxArticleSize > 0 AND PostedSize > CurGroup.MaxArticleSize THEN Oversize = 1
      LOOP UNTIL TempLine = "."
      LogLine("Article received. Running sanity checks...")
      REM The whole article has been retrieved, now let's decide what to do with it...

      REM Check if DstNewsGroups is okay (contains local groups only)
      IF INSTR(DstNewsGroups, ",") > 0 THEN ErrString = "Cross-posting is not supported yet. Sorry."
      IF LEN(ErrString) = 0 THEN
        IF DoesGroupExists(DstNewsGroups) = 0 THEN ErrString = "Unknown newsgroup!"
      END IF

      IF LEN(ErrString) = 0 THEN
        TempGroup = GetGroupProperties(DstNewsGroups)
        REM Check if the group allows posting at all...
        IF TempGroup.WriteAccess = 0 THEN ErrString = "The selected group is write-protected!"
        REM Check if authentication is required for this group (if so, is the user authenticated yet?)
        IF CheckAuth(DstNewsGroups) = 0 THEN ErrString = "This group requires authentication!"
      END IF

      REM Check if FollowupTo is okay
      IF LEN(ErrString) = 0 THEN
        IF LEN(FollowupTo) > 0 THEN ErrString = "The FollowUp-To mechanism is not supported yet. Sorry."
      END IF

      REM Check wheter it is a control message or not
      IF LEN(ErrString) = 0 THEN
        IF LEN(ControlMsg) > 0 THEN ErrString = "Control messages aren't supported yet. Sorry."
      END IF

      REM Check if the message is not bigger then the allowed maximum size
      IF LEN(ErrString) = 0 THEN
        IF OverSize <> 0 THEN ErrString = "You're not allowed to post articles bigger than " & CurGroup.MaxArticleSize & " bytes!"
      END IF

      IF LEN(ErrString) > 0 THEN
          PRINT "441 " + ErrString + CRLF;
          LogLine("Error: " + ErrString)
        ELSE
          TempGroup = GetGroupProperties(DstNewsGroups)
          BodyLines -= 1   ' Do not count the last "." line.
          RightNOW = NowGMT()
          MsgID = "<" & TempGroup.LastArticle + 1 & "-" & TempGroup.Name & "-" & FORMAT(DAY(RightNOW), "00") & FORMAT(MONTH(RightNOW), "00") & FORMAT(YEAR(RightNOW), "0000") & FORMAT(HOUR(RightNOW), "00") & FORMAT(MINUTE(RightNOW), "00") & FORMAT(SECOND(RightNOW), "00") & "-" & HEX(INT(RND * 255), 2) & "@" & ServerFQDN & ">"
          DateReceived = ShortWeekName(WEEKDAY(RightNOW, 2)) & ", " & FORMAT(DAY(RightNOW), "00") & " " & ShortMonth(MONTH(RightNOW)) & " " & YEAR(RightNOW) & " " & FORMAT(HOUR(RightNOW), "00") & ":" & FORMAT(MINUTE(RightNOW), "00") & ":" & FORMAT(SECOND(RightNOW), "00") & " +0000"
          IF LEN(XoverField(4)) = 0 THEN XoverField(4) = DateReceived
          XoverField(1) = "" & TempGroup.LastArticle + 1
          XoverField(5) = MsgID
          XoverField(8) = "" & BodyLines
          XoverField(9) = "Xref: " & ServerFQDN & " " & TempGroup.Name & ":" & TempGroup.LastArticle + 1
          FOR x = 1 TO 9                                 '
            TempLine = ""                                ' Sanitaze all
            FOR y = 1 TO LEN(XoverField(x))              ' fields which
              IF MID(XoverField(x), y, 1) = CHR(9) THEN  ' are destinated
                  TempLine += " "                        ' to go to the
                ELSE                                     ' overview NNTP
                  TempLine += MID(XoverField(x), y, 1)   ' database.
              END IF                                     ' (replace any
            NEXT y                                       ' TAB by a single)
            XoverField(x) = TempLine                     ' space).
          NEXT x                                         '
          FileHandler = FREEFILE
          TempGroup.LastArticle += 1
          LogLine("Saving article on-disk...")
          OPEN SpoolPath & TempGroup.Name & "/" & TempGroup.LastArticle FOR OUTPUT AS #FileHandler
          REM  Add some server's headers
          IF LEN(MsgPath) = 0 THEN MsgPath = ServerFQDN + "!not-for-mail" ELSE MsgPath = ServerFQDN + "!" + MsgPath
          PRINT #FileHandler, "Path: " + MsgPath + CRLF;
          IF LEN(AuthUser) > 0 THEN
              PRINT #FileHandler, "Sender: " + CurUser + CRLF;
          END IF
          PRINT #FileHandler, "Date: " + XoverField(4) + CRLF;
          PRINT #FileHandler, "Lines: " & BodyLines & CRLF;
          PRINT #FileHandler, "Message-ID: " + MsgID + CRLF;
          PRINT #FileHandler, "NNTP-Posting-Host: " + CurHost + CRLF;
          PRINT #FileHandler, "NNTP-Posting-Date: " & DateReceived + CRLF;
          PRINT #FileHandler, XoverField(9) & CRLF; ' Xref header
          REM  Server's headers done.
          FOR x = 1 TO LastHeaderLine
            y = 0
            ' Skip any headers which were already set by the server (prevent injection)
            IF LEFT(UCASE(PostedContent(x)), 7) = "SENDER:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 18) = "NNTP-POSTING-HOST:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 6) = "LINES:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 11) = "MESSAGE-ID:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 18) = "NNTP-POSTING-DATE:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 5) = "DATE:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 5) = "PATH:" THEN y = 1
            IF LEFT(UCASE(PostedContent(x)), 5) = "XREF:" THEN y = 1
            IF y = 0 THEN PRINT #FileHandler, PostedContent(x) + CRLF;
          NEXT x
          REM Headers done, save the body now
          FOR x = LastHeaderLine + 1 TO PostedLines
            IF x < PostedLines OR PostedContent(x) <> "." THEN
              PRINT #FileHandler, PostedContent(x) + CRLF;
            END IF
          NEXT x
          CLOSE #FileHandler
          REM  Update the overview file
          XoverField(7) = "" & GetFileSize(SpoolPath & TempGroup.Name & "/" & TempGroup.LastArticle)
          OPEN SpoolPath & TempGroup.Name & "/overview.idx" FOR APPEND AS #FileHandler
          PRINT #FileHandler, XoverField(1) + CHR(9) + XoverField(2) + CHR(9) + XoverField(3) + CHR(9) + XoverField(4) + CHR(9) + XoverField(5) + CHR(9) + XoverField(6) + CHR(9) + XoverField(7) + CHR(9) + XoverField(8) + CHR(9) + XoverField(9) + CRLF;
          CLOSE #FileHandler
          REM  Overview file updated!
          PRINT "240 Article received OK" + CRLF;
          LogLine("Article saved: " + MsgID)
      END IF
    ELSE
      PRINT "440 Posting not permitted." + CRLF;
      LogLine("Posting not permitted.")
  END IF
END SUB


SUB SwitchMode(RequestMode AS STRING)
  SELECT CASE UCASE(RequestMode)
    CASE "READER"
      IF WriteAccess = 1 THEN
          PRINT "200 Reader mode, posting permitted." + CRLF;
          LogLine("Reader mode OK. Posting allowed.")
        ELSE
          PRINT "201 Reader mode, posting prohibited." + CRLF;
          LogLine("Reader mode OK. Posting not allowed.")
      END IF
    CASE ELSE
      PRINT "501 Unknown MODE option!" + CRLF;
      LogLine("Unknown MODE option.")
  END SELECT
END SUB


SUB AnnounceCapabilities()
  PRINT "101 Capability list:" + CRLF;
  PRINT "VERSION 2" + CRLF;
  PRINT "READER" + CRLF;
  PRINT "LIST ACTIVE NEWSGROUPS OVERVIEW.FMT" + CRLF;
  PRINT "OVER" + CRLF;
  IF WriteAccess = 1 AND CurGroup.WriteAccess = 1 THEN
    PRINT "POST" + CRLF;
  END IF
  IF LEN(AuthUser) = 0 THEN
    PRINT "AUTHINFO USER" + CRLF;
  END IF
  PRINT "IMPLEMENTATION BasicNews " + pVer + CRLF;
  PRINT "." + CRLF;
  LogLine("Announcing capabilities.")
END SUB


SUB AskForAuth(SecondParameter AS STRING, ThirdParameter AS STRING)
  DIM AS STRING GivenUser, GivenPassword, TempLine
  DIM AS INTEGER x, PasswordLessLogin = 0
  IF LEN(AuthUser) > 0 THEN
      LogLine("Authentication denied - already authenticated.")
      SLEEP 500, 1  ' Wait 1/2 of a second... Why not?
      PRINT "502 Already authenticated" + CRLF;
    ELSE
      IF SecondParameter = "USER" AND LEN(ThirdParameter) > 0 THEN
          GivenUser = ThirdParameter
          x = 0          '      Check for password-less logins
          WHILE x < UserBaseSize AND LEN(AuthUser) = 0
            x += 1
            IF UserBase(x) = GivenUser + ":" THEN
              AuthUser = GivenUser
              PasswordLessLogin = 1
            END IF
          WEND
          IF PasswordLessLogin = 1 AND LEN(AuthUser) > 0 THEN
              LogLine("User " + AuthUser + " authenticated successfully (password-less identification).")
              PRINT "281 Authentication accepted. Be my guest!" + CRLF;
            ELSE
              PRINT "381 Gimme your password, now." + CRLF;
              LogLine("Authentication in progress...")
              TempLine = LineInput(4)
              ExplodeCommand(TempLine)
              IF LastCommand(1) = "AUTHINFO" AND LastCommand(2) = "PASS" THEN
                  GivenPassword = LastCommand(3)
                  x = 0
                  WHILE x < UserBaseSize AND LEN(AuthUser) = 0
                    x += 1
                    IF UserBase(x) = GivenUser + ":" + GivenPassword THEN AuthUser = GivenUser
                  WEND
                  IF LEN(AuthUser) = 0 THEN
                      LogLine("Authentication failed (" + GivenUser + ")")
                      SLEEP 2000, 1   ' Add a delay against any possible brute-force attack
                      PRINT "481 Authentication rejected." + CRLF;
                    ELSE
                      PRINT "281 Authentication accepted. Be my guest!" + CRLF;
                      CurUser = AuthUser + "@" + CurHost
                      LogLine("User " + AuthUser + " authenticated successfully.")
                  END IF
                ELSE
                  LogLine("Wrong authentication syntax/sequence (" + GivenUser + ")")
                  SLEEP 2000, 1   ' Make the RFC-unaware bastard waiting a bit
                  PRINT "501 Wrong authentication syntax/sequence." + CRLF;
                  ConnClose()  '  Is it really a good idea to close the connection?
              END IF
            END IF
          ELSEIF SecondParameter = "PASS"  AND LEN(ThirdParameter) THEN
            LogLine("Authentication failed - out of sequence.")
            SLEEP 2000, 1   ' Make the RFC-unaware bastard waiting a bit
            PRINT "482 Authentication rejected (out of sequence)." + CRLF;
            ConnClose()  '  Is it really a good idea to close the connection?
          ELSE
            LogLine("Bad syntax.")
            SLEEP 500, 1
            PRINT "501 Bad syntax. Should be AUTHINFO USER <user> or AUTHINFO PASS <password>." + CRLF;
        END IF
  END IF
END SUB


SUB LoadUserBase()
  DIM AS INTEGER FileHandler
  DIM AS STRING LineBuff
  UserBaseSize = 0
  IF FileExists(SpoolPath & "basicnews.users") <> 0 THEN
    FileHandler = FREEFILE
    OPEN SpoolPath & "basicnews.users" FOR INPUT AS #FileHandler
    WHILE NOT EOF(FileHandler)
      LINE INPUT #FileHandler, LineBuff
      LineBuff = TRIM(LineBuff)
      IF LEN(LineBuff) THEN
        IF LEFT(LineBuff, 1) <> "#" THEN
          IF UserBaseSize = UBOUND(UserBase) THEN REDIM PRESERVE UserBase(1 TO UserBaseSize * 2) AS STRING
          UserBaseSize += 1
          UserBase(UserBaseSize) = LineBuff
        END IF
      END IF
    WEND
    CLOSE #FileHandler
  END IF
END SUB


SUB Over(SpecRange AS STRING)
  DIM AS INTEGER FirstInRange, LastInRange, x, OverviewHandler
  DIM AS STRING LineBuff
  IF IsRangeValid(SpecRange) = 0 AND LEN(SpecRange) > 0 THEN
      IF LEFT(SpecRange, 1) = "<" AND RIGHT(SpecRange, 1) = ">" THEN
          PRINT "503 Retrieval by MsgID not (yet) supported." + CRLF;
          LogLine("OVER by MsgID not yet supported.")
        ELSE
          PRINT "501 Syntax error. OVER requires a number, range or MsgID." + CRLF;
          LogLine("501 Syntax error. OVER requires a number, range or MsgID.")
      END IF
    ELSE
      IF CurGroup.Name = "" THEN
          PRINT "412 No news group current selected." + CRLF;
          LogLine("No newsgroup selected!")
        ELSE
          IF SpecRange = "" THEN
              FirstInRange = CurGroup.CurArticle
              LastInRange = CurGroup.CurArticle
            ELSE
              FirstInRange = RangeLow(SpecRange)
              LastInRange = RangeHigh(SpecRange)
              IF LastInRange = -1 THEN LastInRange = CurGroup.LastArticle
          END IF
          IF FirstInRange < CurGroup.FirstArticle THEN FirstInRange = CurGroup.FirstArticle
          IF LastInRange > CurGroup.LastArticle THEN LastInRange = CurGroup.LastArticle
          IF LastInRange - FirstInRange < 0 THEN FirstInRange = 0 : LastInRange = 0
          IF FirstInRange = 0 OR LastInRange = 0 THEN
              PRINT "423 No article(s) selected." + CRLF;
              LogLine("No article selected!")
            ELSE
              CheckOverviewFile(CurGroup.Name)
              OverviewHandler = FREEFILE
              OPEN SpoolPath & CurGroup.Name & "/overview.idx" FOR INPUT AS #OverviewHandler
              PRINT "224 Overview information follows:" + CRLF;
              LogLine("Overview information returned.")
              WHILE NOT EOF(OverviewHandler)
                LINE INPUT #OverviewHandler, LineBuff
                IF LEN(LineBuff) > 4 THEN
                  x = VAL(MID(LineBuff, 1, INSTR(LineBuff, CHR(9)) - 1))
                  IF x >= FirstInRange AND x <= LastInRange THEN PRINT LineBuff + CRLF; ' Check if Xover range matches
                END IF
              WEND
              CLOSE #OverviewHandler
              PRINT "." + CRLF;
          END IF
      END IF
  END IF
END SUB


SUB Xover(SpecRange AS STRING)
  DIM AS INTEGER FirstInRange, LastInRange, x, OverviewHandler
  DIM AS STRING LineBuff
  IF IsRangeValid(SpecRange) = 0 AND LEN(SpecRange) > 0 THEN
      PRINT "501 Invalid range!" + CRLF;
      LogLine("Invalid range!")
    ELSE
      IF CurGroup.Name = "" THEN
          PRINT "412 No news group current selected." + CRLF;
          LogLine("No newsgroup selected!")
        ELSE
          IF SpecRange = "" THEN
              FirstInRange = CurGroup.CurArticle
              LastInRange = CurGroup.CurArticle
            ELSE
              FirstInRange = RangeLow(SpecRange)
              LastInRange = RangeHigh(SpecRange)
              IF LastInRange = -1 THEN LastInRange = CurGroup.LastArticle
          END IF
          IF FirstInRange < CurGroup.FirstArticle THEN FirstInRange = CurGroup.FirstArticle
          IF LastInRange > CurGroup.LastArticle THEN LastInRange = CurGroup.LastArticle
          IF LastInRange - FirstInRange < 0 THEN FirstInRange = 0 : LastInRange = 0
          IF FirstInRange = 0 OR LastInRange = 0 THEN
              PRINT "420 No article(s) selected." + CRLF;
              LogLine("No article selected!")
            ELSE
              CheckOverviewFile(CurGroup.Name)
              OverviewHandler = FREEFILE
              OPEN SpoolPath & CurGroup.Name & "/overview.idx" FOR INPUT AS #OverviewHandler
              PRINT "224 Overview information follows:" + CRLF;
              LogLine("Overview information returned.")
              WHILE NOT EOF(OverviewHandler)
                LINE INPUT #OverviewHandler, LineBuff
                IF LEN(LineBuff) > 4 THEN
                  x = VAL(MID(LineBuff, 1, INSTR(LineBuff, CHR(9)) - 1))
                  IF x >= FirstInRange AND x <= LastInRange THEN PRINT LineBuff + CRLF; ' Check if range matches
                END IF
              WEND
              CLOSE #OverviewHandler
              PRINT "." + CRLF;
          END IF
      END IF
  END IF
END SUB
